import os.path as osp

import hglib

from cubicweb import Binary, ValidationError
from cubicweb.devtools.testlib import CubicWebTC

from cubes.mercurial_server.testutils import MercurialServerTCMixin


class EntitiesBasicTC(MercurialServerTCMixin, CubicWebTC):
    """ Test features of entities which does not require any mercurial server.
    """

    def test_exec_in_hgadmin_clone(self):
        with self.admin_access.client_cnx() as cnx:
            server_config = cnx.entity_from_eid(self.server_config)
            with server_config.exec_in_hgadmin_clone('test commit msg') as repo:
                filename = osp.join(repo.root(), 'whatever.txt')
                with open(filename, 'a') as fobj:
                    fobj.write('Hello World\n')
                repo.add(filename)

        content = self._hgadmin_file_content('whatever.txt')
        self.assertEqual(content, 'Hello World\n')


    def test_mercurial_server_permission(self):
        with self.admin_access.client_cnx() as cnx:
            access_key = cnx.create_entity('SshPubKey',
                                           data=Binary('42'),
                                           reverse_public_key=cnx.user)

            perm = cnx.create_entity('MercurialServerPermission',
                                     permission_level=u'read',
                                     access_key=access_key,
                                     granted_on=self.admin_user_repo)

            self.assertEqual(perm.server_config.eid, self.server_config)
            self.assertEqual(perm.pubkey.getvalue(), '42')

    def test_mercurial_server_access_key(self):
        with self.admin_access.client_cnx() as cnx:
            access_key = cnx.create_entity('SshPubKey',
                                           data=Binary('42'),
                                           reverse_public_key=cnx.user)

            perm = cnx.create_entity('MercurialServerPermission',
                                     permission_level=u'read',
                                     access_key=access_key,
                                     granted_on=self.admin_user_repo)
            mcs = access_key.mercurial_server_configs
            self.assertListEqual([m.eid for m in mcs], [self.server_config])

    def test_access_conf_generation(self):
        with self.admin_access.client_cnx() as cnx:
            msc = cnx.entity_from_eid(self.server_config)
            access_key = cnx.user.public_key[0]
            self.assertEqual(msc.generate_access_conf(),
                             'write user=cw/%(login)s/%(key_eid)d repo=admin_repo\n'
                             % {'login': cnx.user.login,
                                'key_eid': access_key.eid,
                               })


if __name__ == '__main__':
    from logilab.common.testlib import unittest_main
    unittest_main()



# -*- coding: utf-8 -*-
import os.path as osp

import hglib

from cubicweb import ValidationError, Binary
from cubicweb.devtools.testlib import CubicWebTC

from cubes.mercurial_server.testutils import MercurialServerTCMixin

class HooksBasicTC(MercurialServerTCMixin, CubicWebTC):
    # msc = MercurialServerConfig

    def test_nonascii_login(self):
        with self.admin_access.client_cnx() as cnx:
            user = self.create_user(cnx, u'françois')
            cnx.commit()
            with self.assertRaises(ValidationError):
                cnx.create_entity('SshPubKey',
                                  data=Binary('42'),
                                  reverse_public_key=user)
                cnx.commit()

    def _test_permission_prologue(self, cnx):
        access_key = cnx.user.public_key[0]
        repo = cnx.entity_from_eid(self.admin_user_repo)
        perm = repo.reverse_granted_on[0]
        self._assert_end_of_prologue(cnx, access_key.eid, perm)
        return cnx.entity_from_eid(access_key.eid), cnx.entity_from_eid(perm.eid)

    def _assert_end_of_prologue(self, cnx, access_key, perm):
        server_config = cnx.entity_from_eid(self.server_config)
        # Testing addition
        pubkey = self._hgadmin_file_content('keys', server_config.keys_subdir,
                                            cnx.user().login, str(access_key))
        self.assertEqual(pubkey, 'admin public key')

        # Check access.conf content
        content = self._hgadmin_file_content(server_config.conf_filename)
        self.assertEqual(content,
                         'write user=cw/%(login)s/%(key_eid)d repo=admin_repo\n'
                         % {'login': cnx.user().login,
                            'key_eid': access_key})

    def test_msc_regeneration_service(self):
        with self.admin_access.client_cnx() as cnx:
            access_key, perm = self._test_permission_prologue(cnx)
            access_key = access_key.eid
            perm = perm.eid
            userdesc = cnx.user.format_name_for_mercurial()
            msc = cnx.entity_from_eid(self.server_config)

            with msc.exec_in_hgadmin_clone("pouette", user=userdesc) as hgrepo:
                hgrepo.remove(osp.join(hgrepo.root(), 'keys', msc.keys_subdir))
                filename = osp.join(hgrepo.root(), msc.conf_filename_str)
                with open(filename, 'wb') as access:
                    access.write('')
            cnx.commit()

        with self.admin_access.client_cnx() as cnx:
            cnx.call_service('regenerate_hgadmin_repo', eid=self.server_config)
            self._assert_end_of_prologue(cnx, access_key, perm)

    def test_msc_deletion_ok(self):
        with self.admin_access.client_cnx() as cnx:
            server_config = cnx.entity_from_eid(self.server_config)
            conf_filename = server_config.conf_filename
            keys_dir = server_config.keys_subdir
            access_key, _perm = self._test_permission_prologue(cnx)

            cnx.execute('DELETE MercurialServerConfig X WHERE X eid %s' %
                        self.server_config)
            cnx.commit()

            # Check access.conf content
            content = self._hgadmin_file_content(conf_filename)
            self.assertEqual(content, '\n')

            with self.assertRaises(IOError):
                self._hgadmin_file_content('keys', keys_dir,
                                           cnx.user.login, str(access_key))

    def test_msc_deletion_with_rollback_inside(self):
        """ on deletion failure (i.e.: deletion order is issued
        but the trasnaction fails), the hgadmin repo files should
        be untouched -- though we may observe a pair of deletion commit/backout
        at the mercurial repository level
        """
        with self.admin_access.client_cnx() as cnx:
            access_key, perm = self._test_permission_prologue(cnx)
            access_key = access_key.eid
            perm = perm.eid
            repo = cnx.entity_from_eid(self.server_config).hgadmin_repo
            cnx.execute('DELETE MercurialServerConfig X WHERE X eid %s' %
                        self.server_config)
            cnx.rollback()
            repo.refresh() # need to refresh so local_cache is up to date

        with self.admin_access.client_cnx() as cnx:
            self._assert_end_of_prologue(cnx, access_key, perm)

            # check that a 'backout' commit did happen
            server_config = cnx.entity_from_eid(self.server_config)
            hgrepo = hglib.open(server_config.hgadmin_repopath)
            alllogs = '\n'.join([entry[5] for entry in hgrepo.log()])
            self.assertIn('backout', alllogs)

    def test_permission_AUD(self):
        """ Test Adding, Updating and Deleting """
        with self.admin_access.client_cnx() as cnx:
            access_key, perm = self._test_permission_prologue(cnx)
            # Updating
            perm.cw_set(permission_level=u"write")
            cnx.commit() # Triggers hooks
            # Testing update
            # Check access.conf content
            server_config = cnx.entity_from_eid(self.server_config)
            content = self._hgadmin_file_content(server_config.conf_filename)
            self.assertEqual(content,
                             'write user=cw/%(login)s/%(key_eid)d repo=admin_repo\n'
                             % {'login': cnx.user.login,
                                'key_eid': access_key.eid,
                                })


            # Deleting the permission, thus the access_key relation
            perm.cw_delete()
            cnx.commit() # Triggers hooks
            #Check access.conf content
            content = self._hgadmin_file_content(server_config.conf_filename)
            self.assertEqual(content, '\n')


    def test_accesskey_AUD(self):
        """ Test Adding, Updating and Deleting"""
        with self.admin_access.client_cnx() as cnx:
            access_key, perm = self._test_permission_prologue(cnx)
            # Updating
            access_key.cw_set(data=Binary('24'))
            cnx.commit() # Triggers hooks
            # Testing update
            pubkey = self._hgadmin_file_content('keys', 'cw',
                                                cnx.user.login, str(access_key.eid))
            self.assertEqual(pubkey, '24')

            # Deleting the key, thus the access_key relation
            access_key.cw_delete()
            cnx.commit() # Triggers hooks
            # Testing deletion
            with self.assertRaises(IOError) as exc :
                pubkey = self._hgadmin_file_content('keys', 'cw',
                                                    cnx.user.login, str(access_key.eid))

            # Check access.conf content
            server_config = cnx.entity_from_eid(self.server_config)
            content = self._hgadmin_file_content(server_config.conf_filename)
            self.assertEqual(content, '\n')


    def test_consistent_permissions(self):
        """ Test at most one permissions for any (key, repo) pair"""
        with self.admin_access.client_cnx() as cnx:
            # Adding
            access_key = cnx.create_entity('SshPubKey',
                                           data=Binary('42'),
                                           reverse_public_key=cnx.user)
            perm = cnx.create_entity('MercurialServerPermission',
                                     permission_level=u'read',
                                     access_key=access_key,
                                     granted_on=self.admin_user_repo)
            cnx.commit() # Triggers hooks
            with self.assertRaises(ValidationError) as exc :
                # The relations access_key and repo are inlined ...
                perm = cnx.create_entity('MercurialServerPermission',
                                         permission_level=u'deny',
                                         access_key=access_key,
                                         granted_on=self.admin_user_repo)
                cnx.commit() # ... so are checked immediatly even without commit


if __name__ == '__main__':
    from logilab.common.testlib import unittest_main
    unittest_main()



# -*- coding: utf-8 -*-
# copyright 2011 Florent Cayré (FRANCE), all rights reserved.
# contact florent.cayre@gmail.com
#
# This program is free software: you can redistribute it and/or modify it under
# the terms of the GNU Lesser General Public License as published by the Free
# Software Foundation, either version 2.1 of the License, or (at your option)
# any later version.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License for more
# details.
#
# You should have received a copy of the GNU Lesser General Public License along
# with this program. If not, see <http://www.gnu.org/licenses/>.

"""cubicweb-mercurial-server schema"""

from yams.buildobjs import (EntityType, RelationDefinition, SubjectRelation,
                            String, Bytes, DEFAULT_ATTRPERMS)
from yams.constraints import RegexpConstraint

from cubicweb.schema import RQLConstraint, RRQLExpression, ERQLExpression

from cubes.vcsfile.schema import Repository


Repository.__permissions__['add'] += ('users', ) # XXX Validate this
Repository.__permissions__['update'] += ('owners', ) # XXX Validate this
Repository.__permissions__['delete'] += ('owners', ) # XXX Validate this


class MercurialServerConfig(EntityType):
    __permissions__ = {
        'read':   ('managers', 'users'),
        'add':    ('managers',),
        'update': ('managers', 'owners'),
        'delete': ('managers', 'owners'),
        }
    name = String(required=True, maxsize=128, fulltextindexed=True)
    base_url = String(required=True,
                      description=_('Base URL to access the mercurial server (example: "ssh://hg@my.server")'),
                      constraints=[RegexpConstraint(r'.*[^/]$')])
    hgadmin_path = String(required=True, default='hgadmin',
                          description=_('Path of the mercurial repository holding the access '
                                        'configuration file and the keys (relative to base_url)'),
                          constraints=[RegexpConstraint(r'^\w+$')])
    conf_filename = String(required=True,
                           default='access.conf',
                           description=_('The name of the `access.conf`-like file to be generated.'))
    keys_subdir = String(required=True,
                         default=u'cw',
                         description=_('The subdirectory (relative to "hgadmin/keys") in '
                                       'which public keys will be written.'))


class hosted_by(RelationDefinition):
    """ Tell that the Repository subject is hosted and administered by
    the the MercurialServerConfig object"""
    __permissions__ = {
        'read':   ('managers', 'users'),
        'add':    ('managers', RRQLExpression('U has_update_permission S')),
        'delete': ('managers', RRQLExpression('U has_update_permission S')),
        }
    subject = 'Repository'
    inlined = True
    object = 'MercurialServerConfig'
    cardinality = '?*'
    composite = 'object'

class hgadmin_repository(RelationDefinition):
    """ The VCSRepository representing the hgadmin repo of the mercurial-server """
    __permissions__ = {
        'read':   ('managers', 'users'),
        'add':    ('managers',),
        'delete': ('managers',),
        }

    subject = 'MercurialServerConfig'
    object = 'Repository'
    cardinality = '1?'
    inlined = True
    composite = 'subject'


class SshPubKey(EntityType):
    __permissions__ = {
        'read':   ('managers', 'users'),
        'add':    ('managers', 'users'),
        'update': ('managers', ERQLExpression('U public_key X')),
        'delete': ('managers', ERQLExpression('U public_key X')),
        }
    name = String(maxsize=255)
    data = Bytes(required=True)


class public_key(RelationDefinition):
    __permissions__ = {'read':   ('managers', 'users'),
                     'add':    ('managers', RRQLExpression('S identity U')),
                     'delete': ('managers', RRQLExpression('S identity U')),}
    subject = 'CWUser'
    object = 'SshPubKey'
    description=_('Public key')
    composite='subject'
    cardinality='*1'


class MercurialServerPermission(EntityType):
    """ Represent a reified relation : the permission given by a SSH key on a repository """
    __permissions__ = {
        'read':   ('managers', 'users'),
        'add':    ('managers', 'users'),
        'update': ('managers', ERQLExpression('X granted_on R, R hosted_by MSC, U has_update_permission MSC')),
        'delete': ('managers', ERQLExpression('X granted_on R, R hosted_by MSC, U has_update_permission MSC')),
        }

    _perms_for_repo = {
        'read':   ('managers', 'users'),
        'add':    ('managers', RRQLExpression('O hosted_by MSC, U has_update_permission MSC'),
                               RRQLExpression('O owned_by U')),
        'delete': ('managers', RRQLExpression('O hosted_by MSC, U has_update_permission MSC'),
                               RRQLExpression('O owned_by U')),
        }

    __unique_together__ = [('access_key', 'granted_on')]

    access_key = SubjectRelation('SshPubKey',
                                 description=_('SSH public key identifying who the permission is granted to'),
                                 cardinality='1*',
                                 inlined=True,
                                 composite='object')
    granted_on = SubjectRelation('Repository',
                                 description=_('Mercurial Repository on which the permission is granted'),
                                 cardinality='1*',
                                 inlined=True,
                                 composite='object',
                                 constraints=[RQLConstraint('EXISTS (O hosted_by X)')],
                                 __permissions__ =  _perms_for_repo)
    permission_level = String(vocabulary=(_('deny'), _('read'), _('write'),
                                          _('publish')),
                              internationalizable=True)




# -*- coding: utf-8 -*-
# copyright 2011 Florent Cayré (FRANCE), all rights reserved.
# contact florent.cayre@gmail.com
#
# This program is free software: you can redistribute it and/or modify it under
# the terms of the GNU Lesser General Public License as published by the Free
# Software Foundation, either version 2.1 of the License, or (at your option)
# any later version.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License for more
# details.
#
# You should have received a copy of the GNU Lesser General Public License along
# with this program. If not, see <http://www.gnu.org/licenses/>.

import tempfile
import os, os.path as osp
import shutil
from contextlib import contextmanager

import hglib

from cubicweb import Binary

from cubes.vcsfile import bridge, testutils


class MercurialServerTCMixin(testutils.HGRCMixin):
    debug = False

    @classmethod
    def fs_repopath(cls, reponame):
        return osp.join(cls.datadir, 'repos', reponame)

    @classmethod
    def fs_create_mercurial_repopath(cls, reponame):
        """Create repository path, possibly deleting existing content"""
        repopath = cls.fs_repopath(reponame)
        if osp.isdir(repopath):
            shutil.rmtree(repopath)
        os.makedirs(repopath)
        return repopath

    @classmethod
    def create_mercurial_repo(cls, cnx, reponame, **kwargs):
        """ Create a mercurial repository """
        repopath = cls.fs_create_mercurial_repopath(reponame)
        return cnx.create_entity('Repository',
                                 title=unicode(reponame),
                                 type=u'mercurial',
                                 source_url=u'file://' + repopath,
                                 **kwargs)

    @classmethod
    def fs_create_mercurial_server(cls):
        # mercurial server creates repo directory recursively but this is not
        # the case in this fake env: we need to create receiving directory first
        repopath = cls.fs_create_mercurial_repopath('hgadmin')
        source_url = 'file://' + repopath
        hglib.init(source_url.encode('ascii'))
        return repopath

    @classmethod
    def create_mercurial_server(cls, cnx, name='hgadmin'):
        name = unicode(name)
        return cnx.create_entity('MercurialServerConfig',
                                 name=name,
                                 base_url=u'file://' + cls.datapath('repos'),
                                 hgadmin_path=name)

    def refresh(self):
        with self.repo.internal_cnx() as cnx:
            bridge.import_content(cnx, commitevery=1, raise_on_error=True)

    def setup_database(self):
        '''initialize a fake hgadmin repository'''
        super(MercurialServerTCMixin, self).setup_database()
        with self.admin_access.client_cnx() as cnx:
            self.fs_create_mercurial_server()
            self.server_config = self.create_mercurial_server(cnx).eid
            cnx.commit()
        self.refresh()
        with self.admin_access.client_cnx() as cnx:
            self.admin_pubkey = cnx.create_entity(
                    'SshPubKey', name=u'admin-pubkey',
                    data=Binary('admin public key'),
                    reverse_public_key=cnx.user).eid
            cnx.commit()
            self.admin_user_repo = self.create_mercurial_repo(
                cnx, 'admin_repo', hosted_by=self.server_config).eid
            cnx.commit()
        self.refresh()

    def tearDown(self):
        '''removes the fake hgadmin repository'''
        super(MercurialServerTCMixin, self).tearDown()
        # wipe local repo caches: they get old wrong default push paths
        shutil.rmtree(osp.join(self.config.appdatahome, 'repo_cache'))
        if osp.isdir(self.datapath('repos')):
            if not self.debug:
                shutil.rmtree(self.datapath('repos'), ignore_errors=True)
            else:
                print ("[DEBUG] repositories can be found in %s" %
                       self.datapath('repos'))

    def _hgadmin_file_content(self, *path):
        path = tuple(str(x) for x in path)
        repo = hglib.open(self.fs_repopath('hgadmin'))
        repo.update()
        fullpath = osp.join(repo.root(), *path)
        with open(fullpath, 'r') as f:
            return f.read()



"""cubicweb-mercurial-server entity's classes"""

import sys, os, os.path as osp, shutil, tempfile, time, re
from contextlib import contextmanager
from urllib import quote_plus
from urlparse import urlparse

import hglib

from logilab.common.shellutils import tempdir
from logilab.common.decorators import monkeypatch, cachedproperty

from cubicweb.entities import AnyEntity, authobjs
from cubicweb import ValidationError

from cubes.vcsfile.hooks import clone_to_local_cache
from cubes.vcsfile.entities import Repository


class CWUser(authobjs.CWUser):

    def format_name_for_mercurial(self):
        if self.firstname and self.surname:
            result = u'{firstname} {surname}'.format(
                     firstname=self.firstname,
                     surname=self.surname)
        else:
            result = u'CWuser "{login}"'.format(login=self.login)
        if self.primary_email:
            result += u' <{email}>'.format(email=self.primary_email[0].address)
        return result.encode('utf-8')


class HGSRepository(Repository):

    @contextmanager
    def pull_push(self, message, **commitopts):
        """ Allows edition of a repository between a pull/push
        Use at your own risks.
        """
        if not osp.exists(self.localcachepath):
            clone_to_local_cache(self)
        commitopts.update(message=message)

        tmpdir = tempfile.mkdtemp()
        try:
            repo = hglib.clone(str(self.localcachepath), tmpdir)
            with repo:
                repo.pull(source=self.source_url.encode('ascii'))
                repo.update()
                yield repo
                try:
                    repo.commit(**commitopts)
                except hglib.error.CommandError as exc:
                    if exc.ret == 1 and exc.out.strip() == 'nothing changed':
                        # do not fail if nothing had to be commited
                        return
                    raise
                else:
                    repo.push(dest=self.source_url.encode('ascii'))
        finally:
            shutil.rmtree(tmpdir)

    @property
    def push_url(self):
        if self.source_url.startswith(self.hosted_by[0].base_url + '/'):
            return self.source_url
        # XXX this assumes that the root of the (probably) http source_url is
        # the same as the mercurial-server base_url
        path = urlparse(self.source_url).path
        return self.hosted_by[0].base_url + path


class MercurialServerConfig(AnyEntity):
    __regid__ = 'MercurialServerConfig'

    def dc_long_title(self):
        return u'%s (%s)' % (self.name, self.base_url)

    @property
    def hgadmin_repo(self):
        return self.hgadmin_repository[0]

    @cachedproperty
    def conf_filename_str(self):
        return self.conf_filename.encode(sys.getfilesystemencoding())

    def access_conf_path(self, repo):
        return osp.join(repo.root_str, self.conf_filename)

    @cachedproperty
    def hgadmin_repopath(self):
        return str(self.hgadmin_repository[0].localcachepath)

    @contextmanager
    def exec_in_hgadmin_clone(self, message, **commitopts):
        with self.hgadmin_repo.pull_push(message, **commitopts) as repo:
            yield repo

    def write_access_conf(self, repo, access_line):
        access_conf_path = self.access_conf_path(repo)
        with open(access_conf_path, 'a') as access_file:
            access_file.write(access_line)
        repo.add(access_conf_path)

    def generate_access_conf(self):
        """ Generate the contents of the access.conf file """
        access_conf = []
        for repo in self.reverse_hosted_by:
            assert repo.push_url.startswith(self.base_url + '/')
            repo_path = repo.push_url[len(self.base_url) + 1:]
            for perm in repo.reverse_granted_on:
                key_path = perm.access_key[0].pathtuple
                conf_line = '{level} user={key} repo={repo}'.format(
                            level=perm.permission_level,
                            key=osp.join(self.keys_subdir, *key_path),
                            repo=repo_path)
                access_conf.append(conf_line)
        return '\n'.join(access_conf) + '\n'


class MercurialServerPermission(AnyEntity):
    __regid__ = 'MercurialServerPermission'

    def dc_title(self):
        """Concatenate permission name with user dc_title."""
        try:
            user = self.access_key[0].reverse_public_key[0]
        except IndexError:
            return self._cw._(self.permission_level)
        else:
            return self._cw._('%s on %s for %s') % (self.permission_level, self.granted_on[0].dc_title(), user.dc_title())

    @property
    def pubkey(self):
        return self.access_key[0].data

    @property
    def server_config(self):
        return self.granted_on[0].hosted_by[0]


class SshPubKey(AnyEntity):
    __regid__ = 'SshPubKey'

    def dc_title(self):
        return (self.name or
                self._cw._('%s public key') % self.reverse_public_key[0].dc_title())

    @property
    def mercurial_server_configs(self):
        perms = self.reverse_access_key
        return [perm.server_config for perm in perms]

    @property
    def pathtuple(self):
        """ Implements naming convention <login>/<keyeid>"""
        return  (self.reverse_public_key[0].login.encode('ascii'), str(self.eid))



sync_schema_props_perms('Repository')
sync_schema_props_perms('hosted_by')
sync_schema_props_perms('public_key')

for msc in rql('MercurialServerConfig X').entities():
    if msc.base_url.endswith('/'):
        msc.cw_set(base_url=msc.base_url.rstrip('/'))

sync_schema_props_perms('MercurialServerConfig')



"""mercurial_server user interface

:organization: Logilab
:copyright: 2013 LOGILAB S.A. (Paris, FRANCE), all rights reserved.
:contact: http://www.logilab.fr/ -- mailto:contact@logilab.fr
"""
__docformat__ = "restructuredtext en"
_ = unicode

from cubicweb.predicates import is_instance, match_user_groups, match_form_params, has_related_entities
from cubicweb.view import EntityView
from cubicweb.web import Redirect, RequestError
from cubicweb.web.action import Action
from cubicweb.web.controller import Controller
from cubicweb.web.formwidgets import HiddenInput
from cubicweb.web.views import uicfg
from cubicweb.web.views import ibreadcrumbs

from cubes.vcsfile.views import primary as vcsfile

def setup_ui(vreg):

    def hide(attr, **kwargs):
        kwargs['widget'] = HiddenInput
        uicfg.autoform_field_kwargs.tag_attribute(attr, kwargs)

    afs = uicfg.autoform_section
    pvs = uicfg.primaryview_section
    rct = uicfg.reledit_ctrl
    afs.tag_subject_of(('MercurialServerConfig', 'hgadmin_repository', '*'), 'main', 'hidden')

    # Repository
    afs.tag_subject_of(('Repository', 'hosted_by', '*'), 'main', 'attributes')
    pvs.tag_object_of(('*', 'granted_on', 'Repository'), 'hidden')
    uicfg.autoform_field_kwargs.set_fields_order(
        'Repository', ('title', 'hosted_by', 'source_url'))

    # SshPubKey
    rct.tag_subject_of(('CWUser', 'public_key', '*'), {'reload': True})
    afs.tag_object_of(('CWUser', 'public_key', 'SshPubKey'), 'main', 'hidden')
    afs.tag_attribute(('SshPubKey', 'name'), 'inlined', 'attributes')
    afs.tag_subject_of(('SshPubKey', 'public_key', 'CWUser'), 'main', 'hidden')
    pvs.tag_object_of(('SshPubKey', 'public_key', 'CWUser'), 'attributes')


class MSPIBreadCrumbsAdapter(ibreadcrumbs.IBreadCrumbsAdapter):
    __select__ = is_instance('MercurialServerPermission')
    def parent_entity(self):
        return self.entity.granted_on[0]


class MSCRegenerateAction(Action):
    __regid__ = 'regenerate'
    __select__ = is_instance('MercurialServerConfig') & match_user_groups('managers')

    title = _('regenerate')

    def url(self):
        entity = self.cw_rset.get_entity(0, 0)
        repoeid = entity.eid
        return self._cw.build_url('regenerate_hgadmin_repo', eid=repoeid)


class RegenerateMSC(Controller):
    """Controller used to call the regenerate service"""
    __regid__ = 'regenerate_hgadmin_repo'
    __select__ = Controller.__select__ & match_user_groups('managers') & match_form_params('eid')

    def publish(self, rset=None):
        eid = int(self._cw.form['eid'])
        msc = self._cw.entity_from_eid(eid)
        if msc.e_schema.type != 'MercurialServerConfig':
            raise RequestError('a mercurial server configuration is expected')
        self._cw.cnx.call_service('regenerate_hgadmin_repo', async=False,
                                  eid=msc.eid)
        raise Redirect(msc.absolute_url())


class RepositoryMSPermissionsTab(EntityView):
    __regid__ = _('mercurialserver.repositorypermissions_tab')
    __select__ = is_instance('Repository') & has_related_entities('hosted_by')

    def entity_call(self, entity):
        rset = entity.related('granted_on', role='object')
        self.wview('table', rset, 'null')

vcsfile.RepositoryPrimaryView.tabs.append(RepositoryMSPermissionsTab.__regid__)


def registration_callback(vreg):
    vreg.register_all(globals().values(), __name__)
    setup_ui(vreg)



# -*- coding: utf-8 -*-
import os.path as osp

from cubicweb import Unauthorized, ValidationError, Binary
from cubicweb.devtools.testlib import CubicWebTC

from cubes.mercurial_server.testutils import MercurialServerTCMixin


class SchemaPermTC(MercurialServerTCMixin, CubicWebTC):

    def test_ordinary_user_AUD(self):
        with self.admin_access.client_cnx() as cnx:
            toto = self.create_user(cnx, u'toto').eid
            titi = self.create_user(cnx, u'titi').eid
            cnx.commit()

        with self.new_access('toto').client_cnx() as cnx:
            # toto cannot create a mercurial server config
            with self.assertRaises(Unauthorized):
                self.create_mercurial_server(cnx, 'for_toto')
                cnx.commit()

        with self.new_access('toto').client_cnx() as cnx:
            # without a public key, toto cannot create a hosted repository
            toto_repo = self.create_mercurial_repo(cnx, 'toto_repo')
            with self.assertRaises(ValidationError) as wraperr:
                toto_repo.cw_set(hosted_by=self.server_config)
            self.assertEqual({'hosted_by': 'To create a hosted repository, toto must have a public key'},
                             wraperr.exception.args[1])

        with self.new_access('titi').client_cnx() as cnx:
            # user titi owns a Repository, but not hosted_by
            pubkey = cnx.create_entity('SshPubKey', data=Binary('titi public key'),
                                       reverse_public_key=titi)
            cnx.commit()
            titi_repo = self.create_mercurial_repo(cnx, 'titi_repo')
            cnx.commit()

        with self.new_access('toto').client_cnx() as cnx:
            pubkey = cnx.create_entity('SshPubKey',
                                       data=Binary('toto public key'),
                                       reverse_public_key=toto)
            self.create_mercurial_repo(cnx, 'toto_repo',
                                       hosted_by=self.server_config)
            cnx.commit()
        with self.admin_access.client_cnx() as cnx:
            toto_user = cnx.entity_from_eid(toto)
            toto_key = toto_user.public_key[0].eid
            toto_login = toto_user.login

            conf_filename = cnx.entity_from_eid(self.server_config).conf_filename
        content = self._hgadmin_file_content(conf_filename)
        self.assertListEqual(
            ['write user=cw/admin/%d repo=admin_repo' % self.admin_pubkey,
             'write user=cw/toto/%d repo=toto_repo' % toto_key],
            content.splitlines())
        with self.admin_access.client_cnx() as cnx:
            keys_subdir = cnx.entity_from_eid(self.server_config).keys_subdir
        pk = self._hgadmin_file_content('keys', keys_subdir,
                                        toto_login, str(toto_key))
        self.assertEqual('toto public key', pk)
        with self.admin_access.client_cnx() as cnx:
            cnx.execute('DELETE CWUser U WHERE U login "toto"')
            cnx.commit()
            conf_filename = cnx.entity_from_eid(self.server_config).conf_filename
        content = self._hgadmin_file_content(conf_filename)

        # all info about toto is gone
        self.assertEqual(['write user=cw/admin/%d repo=admin_repo' %
                          self.admin_pubkey],
                         content.splitlines())
        with self.admin_access.client_cnx() as cnx:
            keys_subdir = cnx.entity_from_eid(self.server_config).keys_subdir
        with self.assertRaises(IOError):
            self._hgadmin_file_content('keys', keys_subdir,
                                       toto_login, str(toto_key))

    def test_delete_hosted_by_non_owner(self):
        # just make sure non-owners of a Repository cannot change its
        # hosted_by relation (as per schema permissions)

        with self.admin_access.client_cnx() as cnx:
            toto = self.create_user(cnx, u'toto').eid
            cnx.commit()
        with self.new_access('toto').client_cnx() as cnx:
            repo = cnx.entity_from_eid(self.admin_user_repo)
            self.assertNotIn(toto, set(o.eid for o in repo.owned_by))
            config = repo.hosted_by[0]
            with self.assertRaises(Unauthorized):
                cnx.execute('DELETE R hosted_by C WHERE R eid %d,'
                            '   C eid %d' %
                            (self.admin_user_repo, config.eid))
                cnx.commit()
        with self.admin_access.client_cnx() as cnx:
            repo = cnx.entity_from_eid(self.admin_user_repo)
            repopath = repo.source_url[7:]
            self.assertTrue(osp.exists(repopath))

    def test_repo_deletion_non_owner(self):
        # make sure non-owners cannot delete a repository (as per schema
        # permissions)
        with self.admin_access.client_cnx() as cnx:
            toto = self.create_user(cnx, u'toto')
            cnx.commit()
        with self.new_access('toto').client_cnx() as cnx:
            repo = cnx.entity_from_eid(self.admin_user_repo)
            with self.assertRaises(Unauthorized):
                cnx.execute('DELETE Repository R WHERE R eid %d' %
                            self.admin_user_repo)
                cnx.commit()
        with self.new_access('toto').client_cnx() as cnx:
            repo = cnx.entity_from_eid(self.admin_user_repo)
            repopath = repo.source_url[7:]
            self.assertTrue(osp.exists(repopath))

    def test_msperm_owner(self):
        with self.admin_access.client_cnx() as cnx:
            owner = self.create_user(cnx, u'owner')
            owner.cw_set(reverse_owned_by=self.admin_user_repo)
            cnx.commit()
        with self.new_access(u'owner').client_cnx() as cnx:
            rset = cnx.execute('INSERT SshPubKey K: K data %(d)s, U public_key K WHERE U eid %(u)s',
                               {'u': cnx.user.eid, 'd': Binary(b'toto')})
            cnx.execute('INSERT MercurialServerPermission P: '
                        'P granted_on R, P permission_level "write", '
                        'P access_key K WHERE R eid %(r)s, K eid %(k)s',
                        {'r': self.admin_user_repo, 'k': rset[0][0]})

    def test_msperm_non_owner(self):
        with self.admin_access.client_cnx() as cnx:
            toto = self.create_user(cnx, u'toto')
            cnx.commit()
        with self.new_access('toto').client_cnx() as cnx:
            rset = cnx.execute('INSERT SshPubKey K: K data %(d)s, U public_key K WHERE U eid %(u)s',
                               {'u': cnx.user.eid, 'd': Binary(b'toto')})
            with self.assertRaises(Unauthorized):
                cnx.execute('INSERT MercurialServerPermission P: '
                            'P granted_on R, P permission_level "write", '
                            'P access_key K WHERE R eid %(r)s, K eid %(k)s',
                            {'r': self.admin_user_repo, 'k': rset[0][0]})


if __name__ == '__main__':
    import unittest
    unittest.main()



# -*- coding: utf-8 -*-
# copyright 2011 Florent Cayré (FRANCE), all rights reserved.
# contact florent.cayre@gmail.com
#
# This program is free software: you can redistribute it and/or modify it under
# the terms of the GNU Lesser General Public License as published by the Free
# Software Foundation, either version 2.1 of the License, or (at your option)
# any later version.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License for more
# details.
#
# You should have received a copy of the GNU Lesser General Public License along
# with this program. If not, see <http://www.gnu.org/licenses/>.
"""cubicweb-mercurial-server specific hooks and operations"""

import sys
import os
import os.path as osp
import shutil
import tempfile
from collections import defaultdict

import hglib

from cubicweb import ValidationError
from cubicweb.predicates import is_instance, match_user_groups
from cubicweb.server import Service
from cubicweb.server.hook import Hook, Operation, DataOperationMixIn, match_rtype

# MercurialServerConfig

class MSCCreateHook(Hook):
    """ MSC creation -> automatically create its hgadmin Repository
    Deletion: the mercurial-server data will be out-of sync until
    a new MSC is created again, which will reset its keys & configs.
    """
    __regid__ = 'mercurialserver.config'
    __select__ = Hook.__select__ & is_instance('MercurialServerConfig')
    events = ('before_add_entity',)

    def __call__(self):
        edit = self.entity.cw_edited
        base_url = edit.get('base_url')
        hgadmin_path = edit.get('hgadmin_path')
        if not base_url or not hgadmin_path:
            return # standard validation will catch this, let's bail out
        base_url = edit['base_url'] = base_url.rstrip('/')
        url = '/'.join((base_url, hgadmin_path))
        repo = self._cw.create_entity('Repository',
                                      title=u'%s_repository_for_%s' % (hgadmin_path,
                                                                       self.entity.eid),
                                      type=u'mercurial',
                                      source_url=unicode(url),
                                      has_anonymous_access=False)
        edit['hgadmin_repository'] = repo.eid


class MSCUpdateHook(Hook):
    """strip trailing /s from base_url, if modified"""
    __regid__ = 'mercurialserver.config.update'
    __select__ = Hook.__select__ & is_instance('MercurialServerConfig')
    events = ('before_update_entity',)

    def __call__(self):
        edit = self.entity.cw_edited
        if 'base_url' not in edit:
            return
        edit['base_url'] = edit['base_url'].rstrip('/')


class MercurialServerRepositoryCreation(Hook):
    """ When a Repository is linked to a MercurialServerConfig
    through hosted_by, we try to initialize its remote location
    Failing to do so because the remote repo already exists is not
    a failure.
    """
    __regid__ = 'mercurialserver.initrepo'
    __select__ = Hook.__select__ & match_rtype('hosted_by')
    events = ('after_add_relation', )
    category = 'mercurialserver.fs'

    def __call__(self):
        # we need to refresh user's relations cache and cw_clear_cache is not
        # thread safe while _cw.user is a shared object. Prefer this method.
        user = self._cw.entity_from_eid(self._cw.user.eid)
        repo = self._cw.entity_from_eid(self.eidfrom)
        if not user.public_key:
            msg = (self._cw._('To create a hosted repository, %s must have a public key')
                   % user.login)
            raise ValidationError(self.eidfrom, {'hosted_by': msg})
        try:
            uri = repo.push_url.encode('ascii')
        except UnicodeEncodeError:
            msg = self._cw._('source_url must be ASCII encodable')
            raise ValidationError(self.eidfrom, {'source_url': msg})

        try:
            hglib.init(uri, ssh=self._cw.vreg.config['hg-ssh'] or None)
            self.info('init-ed repository %s', uri)
            # prepare possible rollback (destroy remote repository)
            MSRepoRollbackOnInit.get_instance(self._cw).add_data(uri)
        except hglib.error.CommandError as exc:
            if re.match(r'^mercurial-server: .+ exists$', exc.err.splitlines()[0]):
                self.warning('repo %s seems to be already there: cannot init', uri)
            else:
                raise ValidationError(self.eidfrom, {'source_url': exc.err})
        # let's create a hg server permission right away
        self._cw.create_entity('MercurialServerPermission',
                               access_key=user.public_key,
                               granted_on=self.eidfrom,
                               permission_level=u'write')


class MSRepoRollbackOnInit(DataOperationMixIn, Operation):
    """ If a repo is init-ed per the hook above we may want
    to wipe it on transaction rollback.
    Alas, there is no method to do this currently except asking
    help to people of the managers group.

    Sending them some mail would be nice.
    """

    def rollback_event(self):
        for uri in self.get_data():
            self.critical('Admin: this repository should be wiped %s', uri)


class MSCDeletion(Hook):
    """ On config deletion, we wipe the parts that are under its control
    in the hgadmin repository. We commit and push immediately, before
    we lose track of the hgadmin_repository relation.
    """
    __regid__ = 'mercurialserver.msc_deletion'
    __select__ = Hook.__select__ & is_instance('MercurialServerConfig')
    events = ('before_delete_entity',)
    category = 'mercurialserver.fs'

    def __call__(self):
        msc = self.entity
        msctitle = msc.dc_title().encode('utf-8')
        message = ('server config `%s` (eid %s) deletion' % (msctitle, msc.eid))
        hgadmin = msc.hgadmin_repo

        if not osp.exists(hgadmin.localcachepath):
            clone_to_local_cache(hgadmin)

        tmpdir = tempfile.mkdtemp()
        try:
            hgrepo = hglib.clone(str(hgadmin.localcachepath), tmpdir)
            with hgrepo:
                hgrepo.pull(source=hgadmin.source_url.encode('ascii'))
                hgrepo.update()

                with open(osp.join(hgrepo.root(), msc.conf_filename), 'wb') as conf_file:
                    conf_file.write('\n')
                hgrepo.remove(osp.join(hgrepo.root(), 'keys', msc.keys_subdir))
                cset = hgrepo.commit(message=message)[1]
                hgrepo.push(dest=hgadmin.source_url.encode('ascii'))
        finally:
            shutil.rmtree(tmpdir)

        rbdata = (msc.eid, msc.hgadmin_repo.eid, msctitle, cset)
        MSCRollbackOnDelete.get_instance(self._cw).add_data(rbdata)


class MSCRollbackOnDelete(DataOperationMixIn, Operation):
    """ When an msc deletion fail, we must undo the changes made to
    the hgadmin repository.
    """

    def rollback_event(self):
        for msceid, repoeid, msctitle, cset in self.get_data():
            hgadmin = self.cnx.entity_from_eid(repoeid)
            message = 'backout `%s` (eid %s) deletion' % (msctitle, msceid)
            # the backout in the managed context may **or may not**
            # commit (depending on whether the the backed out
            # changeset has children; see `hg help backout`) AND
            # pull_push will **not** push is the commit is empty, so
            # need an explicit push here...
            with hgadmin.pull_push(message) as hgrepo:
                hgrepo.backout(rev=cset, message=message)
                hgrepo.push(dest=hgadmin.source_url.encode('ascii'))


# SshPubKey

def schedule_key_sync(cnx, event, keyeid):
    key = cnx.entity_from_eid(keyeid)
    if not key.reverse_public_key and event == 'delete':
        # user has been deleted
        return
    for conf in key.mercurial_server_configs:
        SynchronizeAccesKeysOperation.get_instance(cnx).add_data(
            (event, conf.eid, osp.join('keys', conf.keys_subdir, *key.pathtuple),
             key.data.getvalue()))

class CWUSerDeletion(Hook):
    __regid__ = 'mercurialserver.cwuser_deletion'
    __select__ = Hook.__select__ & is_instance('CWUser')
    events = ('before_delete_entity',)

    def __call__(self):
        # we can't destroy remote repositories but we'll wipe the keys at least
        for key in self.entity.public_key:
            CWUserDeleteOp.get_instance(self._cw)

class CWUserDeleteOp(DataOperationMixIn, Operation):

    def postcommit_event(self):
        with self.cnx.repo.internal_cnx() as cnx:
            for msceid, in cnx.execute('MercurialServerConfig M').rows:
                cnx.call_service('regenerate_hgadmin_repo', eid=msceid,
                                 userdesc=self.cnx.user.format_name_for_mercurial())
            cnx.commit()

class SshPubKeyUpdate(Hook):
    __regid__ = 'mercurialserver.accesskeyupdatehook'
    __select__ = Hook.__select__ & is_instance('SshPubKey')
    events = ('after_update_entity',  # all relations are set
              'before_delete_entity') # all relations are _still_ set

    def __call__(self):
        event = 'add' if 'delete' not in self.event else 'delete'
        schedule_key_sync(self._cw, event, self.entity.eid)


class SshPubKeyLinkUnlink(Hook):
    """ Schedule an update operation for the access key on update.
    All MercurialServerConfigurations using this key must be updated.
    """
    __regid__ = 'mercurialserver.accesskeylinkunlink'
    __select__ = Hook.__select__ & match_rtype('public_key', 'access_key')
    events = ('after_add_relation',)     # key just got created
              # 'before_delete_relation') # key just got deleted

    def __call__(self):
        permeid = self.eidto
        event = 'add' if 'add' in self.event else 'delete'
        if event == 'add':
            # wait till everything is in place
            NewAccessKey.get_instance(self._cw).add_data((event, permeid))
        else:
            # do it asap
            schedule_key_sync(self._cw, event, permeid)


class NewAccessKey(DataOperationMixIn, Operation):

    def precommit_event(self):
        for event, keyeid in self.get_data():
            schedule_key_sync(self.cnx, event, keyeid)


# Permission

class MercurialServerPermission(Hook):
    """Schedule a re-generation operation for the access conf
    The permission level (read, write, ...) has been changed.
    """
    __regid__ = 'mercurialserver.permissionupdatehook'
    __select__ = Hook.__select__ & is_instance('MercurialServerPermission')
    events = ('after_update_entity', 'before_delete_entity')

    def __call__(self):
        if 'delete' in self.event:
            if (not self.entity.granted_on
                or self._cw.deleted_in_transaction(self.entity.granted_on[0].eid)):
                # we are being wiped
                return
        RegenerateAccesConfOperation.get_instance(self._cw).add_data(self.entity.server_config.eid)


# Operations

def remove_key(cnx, repo, keypath):
    filename = osp.join(repo.root(), keypath)
    cnx.info('remove public key %s', filename)
    repo.remove(filename)

def add_key(cnx, repo, keypath, pubkey):
    filename = osp.join(repo.root(), keypath)
    dirpath = osp.dirname(filename)
    if not osp.isdir(dirpath):
        os.makedirs(dirpath)
    with open(filename, 'wb') as keyfile:
        keyfile.write(pubkey)
        cnx.info('add public key %s', filename)
        repo.add(str(filename))


def sync_access_keys(cnx, server_config, op_path_pubkey):
    msg = ('Access keys sync from cube mercurial-server for %s' %
           ', '.join([x[1] for x in op_path_pubkey]))
    userdesc = cnx.user.format_name_for_mercurial()
    with server_config.exec_in_hgadmin_clone(msg, user=userdesc) as repo:
        for op, path, pubkey in op_path_pubkey:
            if op == 'add':
                add_key(cnx, repo, path, pubkey)
            else:
                remove_key(cnx, repo, path)


class SynchronizeAccesKeysOperation(DataOperationMixIn, Operation):
    """ Regenerate the mercurial-server `keys` directory

    add_data takes a tuple of (msc.eid, keypath, pubkey)

    key_path is a tuple representing the path associated to the key
    within the `keys` directory of the `hgadmin` repository.
    """

    def precommit_event(self):
        # Group the operations by MercurialServerConfig
        by_config = defaultdict(list)
        for operation, config_eid, path, pubkey in self.get_data():
            by_config[config_eid].append((operation, path, pubkey))

        for configeid, op_path_pubkey in by_config.iteritems():
            RegenerateAccesConfOperation.get_instance(self.cnx).add_data(config_eid)
            server_config = self.cnx.entity_from_eid(configeid)
            sync_access_keys(self.cnx, server_config, op_path_pubkey)

def write_access_conf(cnx, msc):
    msg = ('Regenerate access configuration %s from cube mercurial-server' %
           msc.dc_title())
    with msc.exec_in_hgadmin_clone(msg) as repo:
        access_conf_contents = msc.generate_access_conf()
        filename = osp.join(repo.root(), msc.conf_filename_str)
        already_exists = osp.exists(filename)
        with open(filename, 'w') as fobj:
            fobj.write(access_conf_contents)
        if not already_exists:
            cnx.info('write %s', filename)
            repo.add(filename)

class RegenerateAccesConfOperation(DataOperationMixIn, Operation):
    """ Regenerate the mercurial-server <access.conf> file"""

    def precommit_event(self):
        for configeid in self.get_data():
            if self.cnx.deleted_in_transaction(configeid):
                continue
            server_config = self.cnx.entity_from_eid(configeid)
            write_access_conf(self.cnx, server_config)



# Misc.

class CheckAsciiLoginFromRtype(Hook):
    __regid__ = 'mercurial_server.check_ascii_login_from_rtype'
    __select__ = Hook.__select__ & match_rtype('public_key')
    events = ('after_add_relation',)

    def __call__(self):
        validate_ascii_login(self._cw, self.eidto, self.eidfrom)


def validate_ascii_login(cnx, msaeid, usereid):
    login = cnx.execute('Any L WHERE U login L, U eid %(u)s',
                        {'u': usereid}).rows[0][0]
    try:
        login.encode('ascii')
    except UnicodeEncodeError:
        msg = _('The user login must be ascii encodable (%s is not).') % login
        raise ValidationError(msaeid, {'public_key': msg})


class RegenerateMSCRepoService(Service):
    __regid__ = 'regenerate_hgadmin_repo'
    __select__ = Service.__select__ & match_user_groups('managers')

    def call(self, eid, userdesc=None):
        msc = self._cw.entity_from_eid(eid)
        # first, clean
        msg = "Regenerate cfg and keys"
        if userdesc is None:
            userdesc = self._cw.user.format_name_for_mercurial()
        with msc.exec_in_hgadmin_clone(msg, user=userdesc) as hgrepo:
            hgrepo.remove(osp.join(hgrepo.root(), 'keys', msc.keys_subdir))
            RegenerateAccesConfOperation.get_instance(self._cw).add_data(eid)
            rset = self._cw.execute('Any K WHERE K is SshPubKey, P access_key K, '
                                    'P granted_on R, R hosted_by M, M eid %(eid)s',
                                    {'eid':eid})
            for key in rset.entities():
                filename = osp.join(hgrepo.root(), 'keys', msc.keys_subdir, *key.pathtuple)
                dirpath = osp.dirname(filename)
                if not osp.isdir(dirpath):
                    os.makedirs(dirpath)
                with open(filename, 'wb') as keyfile:
                    keyfile.write(key.data.getvalue())
                hgrepo.add(str(filename))
        # ... just in case, we don't really know who is responsible for this
        self._cw.commit()



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# pylint: disable=W0142,W0403,W0404,W0613,W0622,W0622,W0704,R0904,C0103,E0611
#
# copyright 2011 Florent Cayré (FRANCE), all rights reserved.
# contact florent.cayre@gmail.com
#
# This program is free software: you can redistribute it and/or modify it under
# the terms of the GNU Lesser General Public License as published by the Free
# Software Foundation, either version 2.1 of the License, or (at your option)
# any later version.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License for more
# details.
#
# You should have received a copy of the GNU Lesser General Public License along
# with this program. If not, see <http://www.gnu.org/licenses/>.

"""Generic Setup script, takes package info from __pkginfo__.py file
"""
__docformat__ = "restructuredtext en"

import os
import sys
import shutil
from os.path import isdir, exists, join, walk

try:
    if os.environ.get('NO_SETUPTOOLS'):
        raise ImportError() # do as there is no setuptools
    from setuptools import setup
    from setuptools.command import install_lib
    USE_SETUPTOOLS = True
except ImportError:
    from distutils.core import setup
    from distutils.command import install_lib
    USE_SETUPTOOLS = False
from distutils.command import install_data

# import required features
from __pkginfo__ import modname, version, license, description, web, \
     author, author_email

if exists('README'):
    long_description = file('README').read()
else:
    long_description = ''

# import optional features
import __pkginfo__
if USE_SETUPTOOLS:
    requires = {}
    for entry in ("__depends__",): # "__recommends__"):
        requires.update(getattr(__pkginfo__, entry, {}))
    install_requires = [("%s %s" % (d, v and v or "")).strip()
                       for d, v in requires.iteritems()]
else:
    install_requires = []

distname = getattr(__pkginfo__, 'distname', modname)
scripts = getattr(__pkginfo__, 'scripts', ())
include_dirs = getattr(__pkginfo__, 'include_dirs', ())
data_files = getattr(__pkginfo__, 'data_files', None)
ext_modules = getattr(__pkginfo__, 'ext_modules', None)
dependency_links = getattr(__pkginfo__, 'dependency_links', ())

BASE_BLACKLIST = ('CVS', '.svn', '.hg', 'debian', 'dist', 'build')
IGNORED_EXTENSIONS = ('.pyc', '.pyo', '.elc', '~')


def ensure_scripts(linux_scripts):
    """
    Creates the proper script names required for each platform
    (taken from 4Suite)
    """
    from distutils import util
    if util.get_platform()[:3] == 'win':
        scripts_ = [script + '.bat' for script in linux_scripts]
    else:
        scripts_ = linux_scripts
    return scripts_

def export(from_dir, to_dir,
           blacklist=BASE_BLACKLIST,
           ignore_ext=IGNORED_EXTENSIONS,
           verbose=True):
    """make a mirror of from_dir in to_dir, omitting directories and files
    listed in the black list
    """
    def make_mirror(arg, directory, fnames):
        """walk handler"""
        for norecurs in blacklist:
            try:
                fnames.remove(norecurs)
            except ValueError:
                pass
        for filename in fnames:
            # don't include binary files
            if filename[-4:] in ignore_ext:
                continue
            if filename[-1] == '~':
                continue
            src = join(directory, filename)
            dest = to_dir + src[len(from_dir):]
            if verbose:
                print >> sys.stderr, src, '->', dest
            if os.path.isdir(src):
                if not exists(dest):
                    os.mkdir(dest)
            else:
                if exists(dest):
                    os.remove(dest)
                shutil.copy2(src, dest)
    try:
        os.mkdir(to_dir)
    except OSError, ex:
        # file exists ?
        import errno
        if ex.errno != errno.EEXIST:
            raise
    walk(from_dir, make_mirror, None)


class MyInstallLib(install_lib.install_lib):
    """extend install_lib command to handle  package __init__.py and
    include_dirs variable if necessary
    """
    def run(self):
        """overridden from install_lib class"""
        install_lib.install_lib.run(self)
        # manually install included directories if any
        if include_dirs:
            base = modname
            for directory in include_dirs:
                dest = join(self.install_dir, base, directory)
                export(directory, dest, verbose=False)

# re-enable copying data files in sys.prefix
old_install_data = install_data.install_data
if USE_SETUPTOOLS:
    # overwrite InstallData to use sys.prefix instead of the egg directory
    class MyInstallData(old_install_data):
        """A class that manages data files installation"""
        def run(self):
            _old_install_dir = self.install_dir
            if self.install_dir.endswith('egg'):
                self.install_dir = sys.prefix
            old_install_data.run(self)
            self.install_dir = _old_install_dir
    try:
        import setuptools.command.easy_install # only if easy_install avaible
        # monkey patch: Crack SandboxViolation verification
        from setuptools.sandbox import DirectorySandbox as DS
        old_ok = DS._ok
        def _ok(self, path):
            """Return True if ``path`` can be written during installation."""
            out = old_ok(self, path) # here for side effect from setuptools
            realpath = os.path.normcase(os.path.realpath(path))
            allowed_path = os.path.normcase(sys.prefix)
            if realpath.startswith(allowed_path):
                out = True
            return out
        DS._ok = _ok
    except ImportError:
        pass

def install(**kwargs):
    """setup entry point"""
    if USE_SETUPTOOLS:
        if '--force-manifest' in sys.argv:
            sys.argv.remove('--force-manifest')
    # install-layout option was introduced in 2.5.3-1~exp1
    elif sys.version_info < (2, 5, 4) and '--install-layout=deb' in sys.argv:
        sys.argv.remove('--install-layout=deb')
    cmdclass = {'install_lib': MyInstallLib}
    if USE_SETUPTOOLS:
        kwargs['install_requires'] = install_requires
        kwargs['dependency_links'] = dependency_links
        kwargs['zip_safe'] = False
        cmdclass['install_data'] = MyInstallData

    return setup(name = distname,
                 version = version,
                 license = license,
                 description = description,
                 long_description = long_description,
                 author = author,
                 author_email = author_email,
                 url = web,
                 scripts = ensure_scripts(scripts),
                 data_files = data_files,
                 ext_modules = ext_modules,
                 cmdclass = cmdclass,
                 **kwargs
                 )

if __name__ == '__main__' :
    install()
